clear all; clc

% Settings ================================================================

%% Parameters of the MAPK system 

pars = [1e-5,  ... EGF_stimulation  (bifurcation parameter; unit is 5e-5 pg/ml)
               ...
          1,   ... ERKpp_SOS1_FB    (bifurcation parameter)
          1,   ... ERKpp_MEK_RAF_FB (bifurcation parameter, shared by 2 fbs)
               ...
         3e5,  ... EGFR_tot
         6e4,  ... Ras_tot
         1e5,  ... Sos1_tot
         6e3,  ... RasGAP_tot
         5e5,  ... Raf_tot
         2e5,  ... MEK1_tot
         3e6,  ... ERK_tot
               ...
         1e-2, ... d1
         1e-5, ... b1
         1e-2, ... u1a
         1e+2, ... u1b
         1e-6, ... b2a
          1  , ... u2a
         1e-7, ... b2b
          1  , ... u2b
         1e-4, ... k2a
         1e-5, ... k2b
         1e-5, ... b3
         1e-2, ... u3
         1e+2, ... k3
         1e-7, ... a2
         1e-2, ... d2
         1e-7, ... p1
         1e-2, ... q1
         3e-6, ... p2
         3e-9, ... p3
         6e-10,... p4
         6e-10,... p6
         1e-2, ... q2
         3e-4, ... q3
         3e-4, ... q4
         1e+2, ... q5
         3e-4  ... q6
       ];

n_coords = 23;

   

%% Parametrs for continuation algorithms

init; 
global cds lds; % unused, but potentially useful

cont_opts = contset;

cont_opts = contset(cont_opts,'InitStepsize',100);
cont_opts = contset(cont_opts,'MinStepsize',  10);
cont_opts = contset(cont_opts,'MaxStepsize',1000);

cont_opts = contset(cont_opts,'MaxCorrIters',  10);
cont_opts = contset(cont_opts,'MaxNewtonIters',25);
cont_opts = contset(cont_opts,'MaxTestIters',  20);
cont_opts = contset(cont_opts,'FunTolerance', 1e-3);
cont_opts = contset(cont_opts,'VarTolerance', 1e-3);
cont_opts = contset(cont_opts,'TestTolerance',1e-1);
cont_opts = contset(cont_opts,'Adapt',1);

cont_opts = contset(cont_opts,'Eigenvalues',  1);
cont_opts = contset(cont_opts,'Singularities',1);


%% Bifurcation analysis - ranges of bifurcation parameters

erkpp_sos1_fb_level_min = 0;
erkpp_sos1_fb_level_max = 2;
erkpp_sos1_fb_n_steps   = 5;  % some lines may be removed in post-processing (trimming)
erkpp_sos1_fb_n_steps_osc = 9;

egf_stimul_level_min = 1e-5;
egf_stimul_level_max = 1e-1;
egf_stimul_n_steps   = 9;  % some linesmay be removed in post-processing (trimming)
egf_stimul_n_steps_osc = 16;


%% Bifurcation analysis in EGF_stimulation parameter -- simulations

cont_opts = contset(cont_opts,'MaxNumPoints',2640);

X = [];

fbs = erkpp_sos1_fb_level_min ...
    :(erkpp_sos1_fb_level_max-erkpp_sos1_fb_level_min)/(erkpp_sos1_fb_n_steps-1) ...
    : erkpp_sos1_fb_level_max;

parfor fbi = 1:length(fbs)
  fb = fbs(fbi);
  display(sprintf('===== ERKpp_Sos1_fb strength: %.2f =====',fb));
    
  display('------------- equlibration -------------');
  ps = pars;  ps(2) = fb;
  model = MAPK_short_fbs_2in1;  % global 
  eq_opts = [];
  time_span = 0:1e2:1e6;
  init_concs = zeros(1,n_coords);
  [t,xeq] = ode23tb(model{2},time_span,init_concs,[], ...
      ps( 1),ps( 2),ps( 3),ps( 4),ps( 5),ps( 6),ps( 7),ps( 8),...
      ps( 9),ps(10),ps(11),ps(12),ps(13),ps(14),ps(15),ps(16),...
      ps(17),ps(18),ps(19),ps(20),ps(21),ps(22),ps(23),ps(24),...
      ps(25),ps(26),ps(27),ps(28),ps(29),ps(30),ps(31),ps(32),...
      ps(33),ps(34),ps(35),ps(36));
  xeq = xeq(end,:)';
  display('Done.')

  display('------------- continuation -------------');
  act_par = 1;
  x0 = init_EP_EP(@MAPK_short_fbs_2in1,xeq(1:n_coords),ps',act_par);
  [X(fbi).x,~,X(fbi).S,~,X(fbi).F] = cont(@equilibrium,x0,[],cont_opts);
  X(fbi).x(n_coords+2,:) = fb; % at pos n_coords+1 are values of bif. param, i.e., EGF_stimulation
  display('Done.') 
end



%% Bifurcation analysis in ERKpp_Sos1_feedback parameter -- simulations

cont_opts = contset(cont_opts,'MaxNumPoints',1750);

egf_stimuls = exp( linspace( log(egf_stimul_level_min), ...
                             log(egf_stimul_level_max), egf_stimul_n_steps ) );
             
xeqs = zeros(length(egf_stimuls),n_coords+2);
last_curve_idx = erkpp_sos1_fb_n_steps;
for egfi = 1:length(egf_stimuls)
  egf_stimul_trial = egf_stimuls(egfi);
  nearest_idx = find(X(1).x(n_coords+1,:) >= egf_stimul_trial, 1)
  xeqs(egfi,:) = X(1).x(:,nearest_idx);
end                     

Y = [];

parfor egfi = 1:length(egf_stimuls)
  egf_stimul = xeqs(egfi,n_coords+1);
  display(sprintf('==== EGF_stimulation level: ~%.6f ===', egf_stimul));

  display('------------ initialization ------------');
  ps = pars;
  ps(1) = egf_stimul;
  ps(2) = 0; % initial sosfb
  display('Done.')

  display('------------- continuation -------------');
  act_par = 2;
  x0 = init_EP_EP(@MAPK_short_fbs_2in1,xeqs(egfi,1:n_coords)',ps',act_par);
  [Y(egfi).x,~,Y(egfi).S,~,Y(egfi).F] = cont(@equilibrium,x0,[],cont_opts);
  display('Done.')
end

for egfi = 1:length(egf_stimuls)
  Y(egfi).x(n_coords+2,:) = Y(egfi).x(n_coords+1,:);
  Y(egfi).x(n_coords+1,:) = xeqs(egfi,n_coords+1); % == egf_stimul === ps(1);
end

X = horzcat(X,Y);

%% Plotting 1/2

% -- Trim to the plot range (commented out, because lines comprising long
%                            steps can be effectively removed)
%
% for xi = 1:erkpp_sos1_fb_n_steps
%   iexcl = find(X(xi).x(n_coords+1,:) > egf_stimul_level_max);
%   X(xi).x(:,iexcl) = [];
%   X(xi).F(:,iexcl) = [];
% end
% %
% for xi = (erkpp_sos1_fb_n_steps+1):(erkpp_sos1_fb_n_steps + egf_stimul_n_steps)
%   iexcl = find(X(xi).x(n_coords+2,3:end) > erkpp_sos1_fb_level_max); % leave at least 2 pts (idx >= 3)
%   X(xi).x(:,iexcl) = [];
%   X(xi).F(:,iexcl) = [];
% end


fig = figure(101);

hopfs = [];
region_borders = [];
for idx = 1:length(X)

  unstable_idcs = find(any(X(idx).F(:,1:end-1) >=0));
  stable_idcs   = find(all(X(idx).F(:,1:end-1) < 0));

  % -- get all Hopf points 
  for label_idx = 1:length(X(idx).S)
    if (strcmp(X(idx).S(label_idx,1).msg,'Hopf') == 1)
      idx_H = X(idx).S(label_idx,1).index;
      
      if idx_H <= length(X(idx).x)
        if (X(idx).x(n_coords+2,idx) >= 0) %????
          hopfs = [hopfs; X(idx).x(n_coords+1,idx_H)...
                          X(idx).x(n_coords+2,idx_H)...
                          X(idx).x(n_coords  ,idx_H)]; % ????
        end
      end
    end
  end
  
  % -- plot continuation path  
  plot3(X(idx).x(n_coords+1,unstable_idcs), ...
        X(idx).x(n_coords+2,unstable_idcs), ...
        X(idx).x(n_coords  ,unstable_idcs), 'c--','LineWidth',2);
  hold on

  if idx <= erkpp_sos1_fb_n_steps
    plot3(X(idx).x(n_coords+1,stable_idcs(find(stable_idcs < unstable_idcs(1)))), ...
          X(idx).x(n_coords+2,stable_idcs(find(stable_idcs < unstable_idcs(1)))), ...
          X(idx).x(n_coords  ,stable_idcs(find(stable_idcs < unstable_idcs(1)))), 'b-','LineWidth',3);
    hold on;
    plot3(X(idx).x(n_coords+1,stable_idcs(find(stable_idcs > unstable_idcs(end)))), ...
          X(idx).x(n_coords+2,stable_idcs(find(stable_idcs > unstable_idcs(end)))), ...
          X(idx).x(n_coords  ,stable_idcs(find(stable_idcs > unstable_idcs(end)))), 'b-','LineWidth',3);
    hold on
  else
    plot3(X(idx).x(n_coords+1,stable_idcs), ...
          X(idx).x(n_coords+2,stable_idcs), ...
          X(idx).x(n_coords  ,stable_idcs), 'b-','LineWidth',3);
    hold on
  end

end


% -- plot Hopf points
plot3(hopfs(:,1),hopfs(:,2),hopfs(:,3),'k*','LineWidth',10);
hold on

grid on;
xlabel('EGF stimulation');  ylabel('Erk_{pp}-Sos1 feedback');  zlabel('ERK_{pp}');
axis( [0 egf_stimul_level_max... 
       0 erkpp_sos1_fb_level_max... 
       0                   3e6] ) % max protein level
set(gca,'xscale','log')



%% Limit cycles

clearvars -except  n_coords  pars  X  hopfs ...
             egf_stimul_level_max     egf_stimul_n_steps     egf_stimul_n_steps_osc ...
          erkpp_sos1_fb_level_max  erkpp_sos1_fb_n_steps  erkpp_sos1_fb_n_steps_osc
          

model = MAPK_short_fbs_2in1;  % model name
time_span = 0:5*60:5*24*60*60;

display('==== LC envelope for ERKpp_Sos1_fb ====');

init_concs = zeros(1,n_coords);
pars(1) = hopfs(end,1);  % EGF_stimul_level
pars(2) = hopfs(end,2);  % ERKppSos1_feedback
erkpp_sos1_fb_level_min = hopfs(end,2);

Xperiods_erkpp_sos1_fb = zeros(erkpp_sos1_fb_n_steps_osc,2);
Xosc_erkppsos1fb       = zeros(erkpp_sos1_fb_n_steps_osc,4);

sosfbs = erkpp_sos1_fb_level_min ...
       :(erkpp_sos1_fb_level_max-erkpp_sos1_fb_level_min)/erkpp_sos1_fb_n_steps_osc ...
       : erkpp_sos1_fb_level_max;

parfor sosfbi = 1:length(sosfbs)
  sosfb = sosfbs(sosfbi);
  ps = pars;
  ps(2) = sosfb;
  display(sprintf('-------- ERKpp_Sos1_fb: %.5f -------',sosfb))
  [ts,traj] = ode23s(model{2},time_span,init_concs,[], ...
      ps( 1),ps( 2),ps( 3),ps( 4),ps( 5),ps( 6),ps( 7),ps( 8),...
      ps( 9),ps(10),ps(11),ps(12),ps(13),ps(14),ps(15),ps(16),...
      ps(17),ps(18),ps(19),ps(20),ps(21),ps(22),ps(23),ps(24),...
      ps(25),ps(26),ps(27),ps(28),ps(29),ps(30),ps(31),ps(32),...
      ps(33),ps(34),ps(35),ps(36));
  Xperiods_erkpp_sos1_fb(sosfbi,:) = [sosfb local_maxima_distance_at_end(ts,traj)];
  since = length(traj(1:end,n_coords))/2;
  xREADOUTsechalf = traj(since:end, n_coords);  % look at the 2nd half of traj
  Xosc_erkppsos1fb(sosfbi,:) = [ps(1) ps(2) min(xREADOUTsechalf) max(xREADOUTsechalf)];
end


display('=== LC envelope for EGF_stimulation ===');
which = erkpp_sos1_fb_n_steps + 1; % arbitrary choice of a curve, last one here
pars(1) = hopfs(which,1);  % EGF_stimul_level
pars(2) = hopfs(which,2);  % ERKppSos1_feedback
egf_stimul_level_min = hopfs(which,1);

Xperiods_egf_stimul = zeros(egf_stimul_n_steps_osc,2);
Xosc_egf_stimul     = zeros(egf_stimul_n_steps_osc,4);
        
egf_stimuls = exp( linspace( log(egf_stimul_level_min), ...
                             log(egf_stimul_level_max), egf_stimul_n_steps_osc ) );
        
parfor egfi = 1:length(egf_stimuls)
  egf_stimul_level = egf_stimuls(egfi);
  ps = pars;
  ps(1) = egf_stimul_level;
  display(sprintf('------ EGF_stimulation: %.5f -------',egf_stimul_level))
  [ts,traj] = ode23s(model{2},time_span,init_concs,[], ...
      ps( 1),ps( 2),ps( 3),ps( 4),ps( 5),ps( 6),ps( 7),ps( 8),...
      ps( 9),ps(10),ps(11),ps(12),ps(13),ps(14),ps(15),ps(16),...
      ps(17),ps(18),ps(19),ps(20),ps(21),ps(22),ps(23),ps(24),...
      ps(25),ps(26),ps(27),ps(28),ps(29),ps(30),ps(31),ps(32),...
      ps(33),ps(34),ps(35),ps(36));
  Xperiods_egf_stimul(egfi,:) = [egf_stimul_level local_maxima_distance_at_end(ts,traj)];
  since = length(traj(:,n_coords))/2;
  xREADOUTsechalf = traj(since:end, n_coords);  % look at the 2nd half of traj
  Xosc_egf_stimul(egfi,:) = [ps(1) ps(2) min(xREADOUTsechalf) max(xREADOUTsechalf)];
end


%% Plotting 2/2

% min,max in the oscillatory region, ERKpp_Sos1_feedback  
plot3(Xosc_erkppsos1fb(:,1),Xosc_erkppsos1fb(:,2),Xosc_erkppsos1fb(:,3),'b.','MarkerSize',12); hold on;
plot3(Xosc_erkppsos1fb(:,1),Xosc_erkppsos1fb(:,2),Xosc_erkppsos1fb(:,4),'b.','MarkerSize',12); hold on

% min,max in the oscillatory region, EGF_stimulation   
plot3(Xosc_egf_stimul(:,1),Xosc_egf_stimul(:,2),Xosc_egf_stimul(:,3),'b.','MarkerSize',12); hold on;
plot3(Xosc_egf_stimul(:,1),Xosc_egf_stimul(:,2),Xosc_egf_stimul(:,4),'b.','MarkerSize',12); hold on
